# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import sys
from typing import Any, Dict, List, Type, TypeVar, cast, IO, Union

if sys.version_info[:2] >= (3, 9):
    # pyre-fixme[21]: Could not find name `_TypedDictMeta` in `typing`.
    from typing import _TypedDictMeta
else:
    # pyre-fixme[21]: Could not find name `_TypedDictMeta` in `typing_extensions`.
    from typing_extensions import _TypedDictMeta
from typing_inspect import get_origin, is_optional_type


if sys.version_info[:2] < (3, 7):
    from typing_inspect import get_last_args as get_args
else:
    from typing_inspect import get_args as get_args


class InvalidJson(json.JSONDecodeError):
    def __init__(self, message: str) -> None:
        super().__init__(message, "", 0)


def _is_primitive(target_type: Type[object]) -> bool:
    return target_type in (int, float, str, bool)


def _is_list(target_type: Type[object]) -> bool:
    return get_origin(target_type) in (List, list)


def _is_dictionary(target_type: Type[object]) -> bool:
    return get_origin(target_type) in (Dict, dict)


def _is_typed_dictionary(target_type: Type[object]) -> bool:
    return isinstance(target_type, _TypedDictMeta)  # pyre-ignore: private API.


def _validate_list(value: object, target_type: Type[List[object]]) -> None:
    if not isinstance(value, list):
        raise InvalidJson(f"`{value}` is not a list")
    (element_type,) = get_args(target_type)
    for element in value:
        _validate_value(element, element_type)


def _validate_dictionary(
    value: object, target_type: Type[Dict[object, object]]
) -> None:
    if not isinstance(value, dict):
        raise InvalidJson(f"`{value}` is not a dictionary")

    key_type, value_type = get_args(target_type)
    for key, value in value.items():
        _validate_value(key, key_type)
        _validate_value(value, value_type)


def _validate_typed_dictionary(value: object, target_type: Type[object]) -> None:
    if not isinstance(value, dict):
        raise InvalidJson(f"`{value}` is not a dictionary")

    for key, value_type in target_type.__annotations__.items():
        if key not in value:
            raise InvalidJson(
                f"{value} of TypedDict {target_type} must contain key {key}"
            )
        _validate_value(value[key], value_type)


def _validate_value(value: object, target_type: Type[object]) -> None:
    if target_type is Any:
        return
    elif _is_list(target_type):
        _validate_list(value, cast(Type[List[object]], target_type))
    elif _is_dictionary(target_type):
        _validate_dictionary(value, cast(Type[Dict[object, object]], target_type))
    elif _is_typed_dictionary(target_type):
        _validate_typed_dictionary(value, target_type)
    elif is_optional_type(target_type):
        if value is None:
            return
        _validate_value(value, get_args(target_type)[0])
    else:
        if target_type not in [int, float, str, bool]:
            raise InvalidJson(f"Invalid value type {target_type}")
        if not isinstance(value, target_type):
            raise InvalidJson(f"`{value}` is not a {target_type}")


T = TypeVar("T")


def load(
    input_: Union[IO[str], IO[bytes]],
    target: Type[T],
    *,
    validate: bool = True,
) -> T:
    return loads(input_.read(), target, validate=validate)


def loads(input_: Union[str, bytes], target: Type[T], *, validate: bool = True) -> T:
    try:
        parsed = json.loads(input_)
        if validate:
            _validate_value(parsed, target)
        return parsed
    except Exception as exception:
        raise InvalidJson(str(exception))


def validate(input: object, target: Type[T]) -> T:
    _validate_value(input, target)
    return cast(T, input)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors
from typing import Any, Callable, Dict, List, TypeVar, Union

from typing_extensions import Protocol

# Surfacing `tests` is important for importers to be able to run our tests
# in their own environment
from . import safe_json, tests, type_variable_operators
from .generic import Generic
from .refinement import none_throws, assert_is_instance, safe_cast


__all__ = [
    "assert_is_instance",
    "DecoratorFactory",
    "Generic",
    "JSON",
    "ListVariadic",
    "none_throws",
    "ParameterSpecification",
    "safe_cast",
    "safe_json",
    "tests",
    "type_variable_operators",
    "TypeVarTuple",
    "Unpack",
]


_T = TypeVar("_T")
_R = TypeVar("_R")
TClass = TypeVar("TClass")

JSON = Union[bool, float, int, str, Dict[str, "JSON"], List["JSON"]]


class ParameterSpecification(list):
    """This kind of type variable captures callable parameter specifications
    (known as argspecs in the runtime and inspect library) instead of types,
    allowing the typing of decorators which transform the return type of the
    given callable.
    For example:
        from typing import TypeVar, Callable, List
        from pyre_extensions import ParameterSpecification
        Tparams = ParameterSpecification("Tparams")
        Treturn = TypeVar("Treturn")
        def unwrap(
            f: Callable[Tparams, List[Treturn],
        ) -> Callable[Tparams, Treturn]: ...
        @unwrap
        def foo(x: int, y: str, z: bool = False) -> List[int]:
            return [1, 2, 3]
    decorates foo into a callable that returns int, but still has the same
    parameters, including their names and whether they are required.

    The list inheritance is required for backwards compatibility with the runtime
    implementation for callables, which requires the first argument to be
    a list.

    The args and kwargs properties are used for specifying that a literal definition
    has the same signature as a ParameterSpecification, like:
    def listify(
        f: Callable[TParams, TReturn]
    ) -> Callable[TParams, List[TReturn]]:
        def wrapped( *args: TParams.args, **kwargs: TParams.kwargs):
            return [f(*args, **kwargs)]
    """

    args = object()
    kwargs = object()

    def __init__(self, *args: object, **kwargs: object) -> None:
        pass


def ListVariadic(name) -> object:
    return Any


_A = TypeVar("_A", bound=int)
_B = TypeVar("_B", bound=int)
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_Ts = ListVariadic("_Ts")
_P = ParameterSpecification("_P")


class DecoratorFactory(Protocol):
    """Usable as a return type for simple decorator factories"""

    def __call__(self, __f: Callable[_P, _R]) -> Callable[_P, _R]:
        ...


class Add(Generic[_A, _B], int):
    pass


class Multiply(Generic[_A, _B], int):
    pass


class Subtract(Generic[_A, _B], int):
    pass


class Divide(Generic[_A, _B], int):
    pass


class Length(Generic[_Ts], int):
    pass


class Product(Generic[_Ts], int):
    pass


class TypeVarTuple:
    def __init__(
        self,
        name: str,
    ) -> None:
        pass


class Unpack(Generic[_T]):
    pass


class Broadcast(Generic[_T1, _T2]):
    pass


class BroadcastError(Generic[_T1, _T2]):
    pass


_Rs = TypeVarTuple("_Rs")


class Compose(Generic[_Rs]):
    pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors
import sys
import unittest
from typing import Any, Callable, Iterable, Iterator, List, TypeVar


class BasicTestCase(unittest.TestCase):
    def test_parameter_specification(self) -> None:
        try:
            from .. import ParameterSpecification

            TParams = ParameterSpecification("TParams")
            TReturn = TypeVar("T")

            def listify(
                f: Callable[TParams, TReturn]
            ) -> Callable[TParams, List[TReturn]]:
                def wrapped(*args: TParams.args, **kwargs: TParams.kwargs):
                    return [f(*args, **kwargs)]

                return wrapped

            def foo():
                return 9

            listify(foo)

        except Exception:
            self.fail("ParameterSpecification missing or broken")

    def test_list_variadics(self) -> None:
        try:
            from .. import ListVariadic
            from ..type_variable_operators import Map

            TReturn = TypeVar("T")
            Ts = ListVariadic("Ts")

            def better_map(
                func: Callable[[Ts], TReturn], *args: Map[Iterable, Ts]
            ) -> Iterator[TReturn]:
                return map(func, *args)

        except Exception:
            self.fail("ListVariadics missing or broken")

    def test_none_throws(self) -> None:
        try:
            from .. import none_throws

            none_throws(0)
            none_throws(0, "custom message")
        except Exception:
            self.fail("none_throws missing or broken")

    def test_safe_cast(self) -> None:
        try:
            from .. import safe_cast

            safe_cast(float, 1)
            safe_cast(1, float)
            safe_cast(Any, "string")
        except Exception:
            self.fail("safe_cast should not have runtime implications")

    def test_generic(self) -> None:
        try:
            from typing import TypeVar  # usort: skip wants trailing whitespace
            from .. import Generic, ListVariadic
            from ..type_variable_operators import Concatenate

            # permitted
            class Foo(Generic):
                pass

            # permitted
            class Bar(Generic[int]):
                pass

            # permitted
            class Baz(Generic[42]):
                pass

            Shape = ListVariadic("Shape")
            DType = TypeVar("DType")

            # intended use
            class Tensor(Generic[Concatenate[DType, Shape]]):
                pass

            # not handled in the backend ... yet
            X = Tensor[int, 7, 8, 9]  # noqa

        except Exception:
            self.fail("Generic/GenericMeta/Concatenate missing or broken")

    def test_generic__metaclass_conflict(self) -> None:
        def try_generic_child_class() -> None:
            from abc import ABC, abstractmethod
            from typing import TypeVar

            from .. import Generic

            T1 = TypeVar("T1")
            T2 = TypeVar("T2")

            class Base(ABC):
                @abstractmethod
                def some_method(self) -> None:
                    ...

            # Python complains at runtime if a child class has a metaclass that
            # is not a strict subclass of the base metaclass.
            # There should be no such metaclass conflict when using Pyre's
            # custom Generic in a child class.
            class Child(Base, Generic[T1, T2]):
                def some_method(self) -> None:
                    ...

        try:
            if sys.version_info >= (3, 7):
                try_generic_child_class()
            else:
                # Older versions of Python will raise the above metaclass
                # conflict `TypeError`.
                with self.assertRaises(TypeError):
                    try_generic_child_class()

        except Exception as exception:
            self.fail(
                "Generic/GenericMeta/Concatenate missing or broken: "
                f"Got exception `{exception}`"
            )

    def test_variadic_tuple(self) -> None:
        try:
            from .. import TypeVarTuple, Unpack

            T = TypeVar("T")
            Ts = TypeVarTuple("Ts")

            def apply(f: Callable[[Unpack[Ts]], T], *args: Unpack[Ts]) -> T:
                return f(*args)

        except Exception:
            self.fail("Variadic tuples missing or broken")

    def test_json(self) -> None:
        try:
            from .. import JSON
        except Exception:
            self.fail("JSON missing or broken")

        def test_json(x: JSON) -> None:
            try:
                # pyre-fixme: Pyre should complain about this
                y = x + 1
            except TypeError:
                # TypeError is anticipated for some instance where
                # Pyre would alert us to an issue if it wasn't
                # set to fixme
                pass

            # Pyre should not complain about these
            if isinstance(x, int):
                y = x + 1
            elif isinstance(x, float):
                y = x + 1.1
            elif isinstance(x, bool):
                y = x or True
            elif isinstance(x, str):
                y = x + "hello"
            elif isinstance(x, list):
                y = x + [4]  # noqa
            elif isinstance(x, dict):
                x["key"] = "value"

        test_json(3)
        test_json(3.5)
        test_json("test_string")
        test_json({"test": "dict"})
        test_json(["test_list"])


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors
import sys
from typing import Any


# We export this metaclass because Concatenate depends on it.
class GenericMeta(type):
    def __getitem__(cls, *args) -> Any:
        return cls.__class__(cls.__name__, cls.__bases__, dict(cls.__dict__))


if sys.version_info >= (3, 7):

    class Generic:
        """Pyre's variadic-supporting substitute for `typing.Generic`.

        By using `__class_getitem__`, this avoids a metaclass, which prevents
        ugly metaclass conflicts when a child class is generic and a base class
        has some metaclass."""

        def __class_getitem__(cls, *args: object) -> Any:
            return cls


else:

    class Generic(metaclass=GenericMeta):
        """Pyre's variadic-supporting substitute for `typing.Generic`."""

        pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import safe_json_test, simple_tests


__all__ = ["safe_json_test", "simple_tests"]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe
from typing import Any, Generic, TypeVar

from .generic import GenericMeta


_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")


class ParameterSpecificationComponentMeta(type):
    def __getitem__(cls, __tparams) -> object:
        return Any


class Map(Generic[_T1, _T2]):
    pass


class Concatenate(metaclass=GenericMeta):
    pass


class PositionalArgumentsOf(metaclass=ParameterSpecificationComponentMeta):
    pass


class KeywordArgumentsOf(metaclass=ParameterSpecificationComponentMeta):
    pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-ignore-all-errors

import unittest
from io import StringIO
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from typing_extensions import TypedDict

from .. import safe_json

T = TypeVar("T")


class Movie(TypedDict):
    name: str
    year: int


class MovieWithRating(Movie):
    rating: float


class MovieWithArbitraryDictionary(Movie):
    dictionary: Dict[str, Any]


class MovieWithUnion(Movie):
    int_or_str: Union[int, str]


class MovieWithNonRequiredField(Movie, total=False):
    not_required: str


class BasicTestCase(unittest.TestCase):
    def _assert_loads(self, input: str, target_type: Type[T], output: T) -> None:
        self.assertEqual(
            safe_json.loads(
                input,
                target_type,
            ),
            output,
        )

    def _assert_loads_fails(self, input: str, target_type: Type[T]) -> None:
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads(
                input,
                target_type,
            )

    def test_loads(self) -> None:
        # Primitives.
        self.assertEqual(safe_json.loads("1", int), 1)
        self.assertEqual(safe_json.loads("true", bool), True)
        self.assertEqual(safe_json.loads("1.1", float), 1.1)
        self.assertEqual(safe_json.loads('"string"', str), "string")

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("1", bool)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("1", float)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("1", str)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("true", float)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("true", str)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("1.1", int)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("1.1", bool)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("1.1", str)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("hello", int)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("hello", bool)

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("hello", float)

        # Lists.
        self.assertEqual(safe_json.loads("[]", List[int]), [])
        self.assertEqual(safe_json.loads("[1]", List[int]), [1])
        self.assertEqual(safe_json.loads("[1, 2]", List[int]), [1, 2])

        self.assertEqual(
            safe_json.loads('[{"1": 1}]', List[Dict[str, int]]), [{"1": 1}]
        )

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads("[1, 'string']", List[int])

        # Dictionaries.
        self.assertEqual(safe_json.loads("{}", Dict[int, str]), {})
        self.assertEqual(safe_json.loads('{"1": 1}', Dict[str, int]), {"1": 1})

        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads('{"1": "string"}', Dict[str, int])
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads('{"1": 1, "2": "2"}', Dict[str, int])

        self.assertEqual(
            safe_json.loads('{"1": {"2": 3}}', Dict[str, Dict[str, int]]),
            {"1": {"2": 3}},
        )

        # Typed dictionaries.
        self.assertEqual(
            safe_json.loads('{"name": "The Matrix", "year": 1999}', Movie),
            {"name": "The Matrix", "year": 1999},
        )
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.loads('{"name": "The Matrix", "year": ""}', Movie)

        # Any.
        self.assertEqual(safe_json.loads("[1]", List[Any]), [1])
        self.assertEqual(safe_json.loads('[{"1": 1}]', List[Any]), [{"1": 1}])

        # Optionals.
        self.assertEqual(safe_json.loads("[1]", List[Optional[int]]), [1])
        self.assertEqual(safe_json.loads("[null, 2]", List[Optional[int]]), [None, 2])

        # Validation can be turned off.
        self.assertEqual(safe_json.loads("[1]", List[str], validate=False), [1])

    def test_validate(self) -> None:
        # Lists.
        parsedListStr = ["1", "2"]

        self.assertEqual(safe_json.validate(parsedListStr, List[str]), parsedListStr)
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.validate(parsedListStr, List[int])

        # Dictionaries.
        parsedDictBasic = {"1": 1}

        self.assertEqual(
            safe_json.validate(parsedDictBasic, Dict[str, int]), parsedDictBasic
        )
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.validate(parsedDictBasic, List[Any])

        parsedDictNested = {"1": {"2": 3}}

        self.assertEqual(
            safe_json.validate(parsedDictNested, Dict[str, Dict[str, int]]),
            parsedDictNested,
        )
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.validate(parsedDictNested, Dict[str, int])

        # Typed dictionaries.
        parsedDictTyped = {"name": "The Matrix", "year": 1999}
        parsedDictTypedFailing = {"name": "The Matrix", "year": ""}

        self.assertEqual(safe_json.validate(parsedDictTyped, Movie), parsedDictTyped)
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.validate(parsedDictTypedFailing, Movie)

        # Any.
        parsedAny = [{"1": 1}]

        self.assertEqual(safe_json.validate(parsedAny, List[Any]), parsedAny)

        # Optionals.

        parsedOptionals = [2, None, 4]
        self.assertEqual(
            safe_json.validate(parsedOptionals, List[Optional[int]]), parsedOptionals
        )

    def test_load(self) -> None:
        f = StringIO('{"1": {"2": 3}}')

        self.assertEqual(safe_json.load(f, Dict[str, Dict[str, int]]), {"1": {"2": 3}})
        with self.assertRaises(safe_json.InvalidJson):
            safe_json.load(f, Dict[int, Dict[int, int]])

    def test_loads_typed_dictionary(self) -> None:
        # Field that is not present in the TypedDict.
        self._assert_loads(
            '{"name": "The Matrix Reloaded", "year": 1999, "extra_field": "hello"}',
            Movie,
            {"name": "The Matrix Reloaded", "year": 1999, "extra_field": "hello"},
        )

        # TypedDict inheriting from another.
        self._assert_loads(
            '{"name": "The Matrix", "year": 1999, "rating": 9.0}',
            MovieWithRating,
            {"name": "The Matrix", "year": 1999, "rating": 9.0},
        )

        self._assert_loads_fails(
            '{"name": "The Matrix", "year": 1999, "rating": "not a float"}',
            MovieWithRating,
        )

        # TypedDict with a field accepting an arbitrary dictionary.
        self._assert_loads(
            '{"name": "The Matrix", "year": 1999,'
            + ' "dictionary": {"foo": "bar", "baz": {}}}',
            MovieWithArbitraryDictionary,
            {
                "name": "The Matrix",
                "year": 1999,
                "dictionary": {"foo": "bar", "baz": {}},
            },
        )

        self._assert_loads_fails(
            '{"name": "The Matrix", "year": 1999, "dictionary": [1, 2]}',
            MovieWithArbitraryDictionary,
        )

        # TODO(T92804673): Unions are not supported.
        self._assert_loads_fails(
            '{"name": "The Matrix", "year": 1999, "int_or_str": 1}',
            MovieWithUnion,
        )

        self._assert_loads(
            '{"name": "The Matrix", "year": 1999, "not_required": "hello"}',
            MovieWithNonRequiredField,
            {"name": "The Matrix", "year": 1999, "not_required": "hello"},
        )

        # TODO(T92805077): Missing non-required field should not be an error.
        self._assert_loads_fails(
            '{"name": "The Matrix", "year": 1999}',
            MovieWithNonRequiredField,
        )


if __name__ == "__main__":
    unittest.main()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyre-extensions",
    version="0.0.27",
    author="Facebook",
    author_email="pyre@fb.com",
    description="Type system extensions for use with the pyre type checker",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pyre-check.org",
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development',
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    install_requires=["typing-inspect", "typing-extensions"],
)
        


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Helper functions for type refinement or asserts.

from typing import Any, Optional, Type, TypeVar

_T = TypeVar("_T")
_TClass = TypeVar("_TClass")


def none_throws(optional: Optional[_T], message: str = "Unexpected `None`") -> _T:
    """Convert an optional to its value. Raises an `AssertionError` if the
    value is `None`"""
    if optional is None:
        raise AssertionError(message)
    return optional


def assert_is_instance(obj: object, cls: Type[_TClass]) -> _TClass:
    """Assert that the given object is an instance of the given class. Raises a
    `TypeError` if not."""
    if not isinstance(obj, cls):
        raise TypeError(f"obj is not an instance of cls: obj={obj} cls={cls}")
    return obj


# pyre-ignore[2]: Intentional `Any` input type.
def safe_cast(new_type: Type[_T], value: Any) -> _T:
    """safe_cast will change the type checker's inference of x if it was
    already a subtype of what we are casting to, and error otherwise."""
    return value
